/**
 * GMyth Library
 *
 * @file gmyth/gmyth_livetv.c
 * 
 * @brief <p> GMythLiveTV starts a remote TV session with the MythTV backend.
 *
 * Copyright (C) 2006 INdT - Instituto Nokia de Tecnologia.
 * @author Rosfran Lins Borges <rosfran.borges@indt.org.br>
 *
 *//*
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
  
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gmyth_livetv.h"
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gmyth_remote_util.h"
#include "gmyth_tvchain.h"
#include "gmyth_socket.h"
#include "gmyth_debug.h"

#include "gmyth_file_transfer.h"
#include "gmyth_monitor_handler.h"

static void gmyth_livetv_class_init          (GMythLiveTVClass *klass);
static void gmyth_livetv_init                (GMythLiveTV *object);

static void gmyth_livetv_dispose  (GObject *object);
static void gmyth_livetv_finalize (GObject *object);

static gint tvchain_curr_index = -1; 

#define GMYTHTV_RECORDER_HEADER         "QUERY_RECORDER"


G_DEFINE_TYPE(GMythLiveTV, gmyth_livetv, G_TYPE_OBJECT)

static void
gmyth_livetv_class_init (GMythLiveTVClass *klass)
{
	GObjectClass *gobject_class;

	gobject_class = (GObjectClass *) klass;

	gobject_class->dispose  = gmyth_livetv_dispose;
	gobject_class->finalize = gmyth_livetv_finalize;	
}

static void
gmyth_livetv_init (GMythLiveTV *livetv)
{
	livetv->backend_info = NULL;
	livetv->local_hostname = NULL;
	livetv->file_transfer = NULL;
	livetv->setup_done = FALSE;

	livetv->recorder = NULL;
	livetv->tvchain = NULL;
	livetv->proginfo = NULL;
}

static void
gmyth_livetv_dispose  (GObject *object)
{

	G_OBJECT_CLASS (gmyth_livetv_parent_class)->dispose (object);
}

static void
gmyth_livetv_finalize (GObject *object)
{
	g_signal_handlers_destroy (object);

	GMythLiveTV *livetv = GMYTH_LIVETV (object);

	gmyth_debug ("[%s] Finalizing livetv", __FUNCTION__);

	if ( livetv->monitor != NULL ) {
		g_object_unref (livetv->monitor);
		livetv->monitor = NULL;
	}

	if ( livetv->recorder != NULL ) {
		g_object_unref (livetv->recorder);
		livetv->recorder = NULL;
	}

	if ( livetv->tvchain != NULL ) {
		g_object_unref (livetv->tvchain);
		livetv->tvchain = NULL;
	}

	if ( livetv->proginfo != NULL ) {
		g_object_unref (livetv->proginfo);
		livetv->proginfo = NULL;
	}
	
	if ( livetv->file_transfer != NULL ) {
		g_object_unref (livetv->file_transfer);
		livetv->file_transfer = NULL;
	}
	
	if ( livetv->backend_info != NULL ) {
		g_object_unref (livetv->backend_info);
		livetv->backend_info = NULL;
	}

	G_OBJECT_CLASS ( gmyth_livetv_parent_class )->finalize ( object );
}

GMythLiveTV*
gmyth_livetv_new ()
{
	GMythLiveTV *livetv = GMYTH_LIVETV ( g_object_new( GMYTH_LIVETV_TYPE, NULL ) );

	return livetv;
}

static void
gmyth_livetv_monitor_signal_handler( GMythMonitorHandler *monitor, gint msg_code, 
							gchar* message, gpointer livetv )
{
	GMythLiveTV *live_tv = (GMythLiveTV*) livetv;
	
	gmyth_debug( "LIVETV Signal handler ( msg = %s, code = %d )\n", message, msg_code );
	
	if ( NULL == live_tv )
		return;		
	
	switch ( msg_code ) 
	{
		
		case GMYTH_BACKEND_PROGRAM_INFO_CHANGED:
		{
			gmyth_debug( "LIVETV Program Changed event received [ msg = %s ]\n", message );
			/*
			if ( !gmyth_livetv_next_program_chain( live_tv ) )
			{
				gmyth_debug( "Cannot change to the next program chain!" );
			}
			else
			{
			*/
			gmyth_debug( "OK!!! MOVED to the next program chain [%s]!", 
									(gmyth_tvchain_get_id( live_tv->tvchain ))->str );
			/* advertises the FileTransfer about the program info changed */
			g_signal_emit ( live_tv->file_transfer,
             GMYTH_FILE_TRANSFER_GET_CLASS (live_tv->file_transfer)->program_info_changed_handler_signal_id,
             0, /* details */
             msg_code, live_tv );
			/*}*/
			break;
		}
		default:
			break;
		
	}
	
}

gboolean
gmyth_livetv_setup ( GMythLiveTV *livetv, GMythBackendInfo *backend_info )
{
	gboolean res = TRUE;

	GMythSocket *socket = gmyth_socket_new ();
	
	livetv->backend_info = backend_info;

	// FIME: Implement this at gmyth_socket
	res = gmyth_socket_connect_to_backend (socket, livetv->backend_info->hostname,
			livetv->backend_info->port, TRUE);
	if (!res) {
		g_warning ("[%s] LiveTV can not connect to backend", __FUNCTION__);	
		res = FALSE;
		goto error;
	}

	livetv->is_livetv = TRUE;
	
  livetv->monitor = gmyth_monitor_handler_new ( );
  
  res = gmyth_monitor_handler_open (livetv->monitor, livetv->backend_info->hostname, 
  				livetv->backend_info->port );
  
  if ( res == TRUE )
  {
  	gmyth_debug("Connect MythTV Monitor event socket! Trying to start the message handler...");
  	
  	res = gmyth_monitor_handler_start ( livetv->monitor );
  	
  	if (res)
  	{
  		gmyth_debug("MythTV Monitor event socket connected and listening!");
  		g_signal_connect (G_OBJECT (livetv->monitor), "backend-events-handler",
                  (GCallback)gmyth_livetv_monitor_signal_handler,
                  livetv);
  	}
  	else
  	{
  		gmyth_debug("Problems when trying to start MythTV Monitor event socket!");
  		goto error;  		
  	}
  }

	livetv->local_hostname  = gmyth_socket_get_local_hostname ( );

	if ( livetv->local_hostname == NULL ) {
		res = FALSE;
		goto error;
	}

	// Gets the recorder num
	livetv->recorder = remote_request_next_free_recorder (socket, -1);
	gmyth_socket_close_connection (socket);

	if ( livetv->recorder == NULL ) {
		g_warning ("[%s] None remote encoder available", __FUNCTION__);
		res = FALSE;
		goto error;
	}

	// Creates livetv chain handler
	livetv->tvchain = GMYTH_TVCHAIN ( g_object_new(GMYTH_TVCHAIN_TYPE, NULL) );
	gmyth_tvchain_initialize ( livetv->tvchain, livetv->backend_info );

	if ( livetv->tvchain == NULL || livetv->tvchain->tvchain_id == NULL ) {
		res = FALSE;
		goto error;
	}

	// Init remote encoder. Opens its control socket.
	res = gmyth_recorder_setup(livetv->recorder);
	if ( !res ) {
		g_warning ("[%s] Fail while setting remote encoder\n", __FUNCTION__);
		res = FALSE;
		goto error;
	}

#if 0
  gint ch = 1011;
  gint idx = 0;  
  for ( ; idx < 5; idx++ ) {
  	if ( gmyth_recorder_check_channel( livetv->recorder, ch ) ) {
	  	if ( gmyth_recorder_set_channel( livetv->recorder, ch ) ) {
	  		g_print( "[%s] Channel changed!!! [%d].\n", __FUNCTION__, ch );
	  		break;
	  	}
  	}
  	++ch;
  }
#endif

	// Spawn live tv. Uses the socket to send mythprotocol data to start livetv in the backend (remotelly)
	res = gmyth_recorder_spawntv ( livetv->recorder,
			gmyth_tvchain_get_id(livetv->tvchain) );
	if (!res) {
		g_warning ("[%s] Fail while spawn tv\n", __FUNCTION__);
		res = FALSE;
		goto error;
	}

	// Reload all TV chain from Mysql database.
	gmyth_tvchain_reload_all (livetv->tvchain);

	if ( livetv->tvchain == NULL ) {
		res = FALSE;
		goto error;
	}

	// Get program info from database using chanid and starttime
	livetv->proginfo = gmyth_tvchain_get_program_at (livetv->tvchain, tvchain_curr_index++ );
	if ( livetv->proginfo == NULL ) {
		g_warning ("[%s] LiveTV not successfully started.\n", __FUNCTION__ );
		res = FALSE;
		goto error;
	} else {
		gmyth_debug ("[%s] GMythLiveTV: All requests to backend to start TV were OK.\n", __FUNCTION__ );
	}
	
	livetv->setup_done = TRUE;
	
	return res;

error:
	g_print( "[%s] ERROR running LiveTV setup.\n", __FUNCTION__ );

	if ( livetv->local_hostname != NULL ) {
		g_string_free( livetv->local_hostname, TRUE );
		res = FALSE;
	}

	if ( livetv->recorder != NULL ) {
		g_object_unref (livetv->recorder);
		livetv->recorder = NULL;
	}

	if ( livetv->tvchain != NULL ) {
		g_object_unref (livetv->tvchain);
		livetv->tvchain = NULL;
	}

	if ( livetv->proginfo != NULL ) {
		g_object_unref (livetv->proginfo);
		livetv->proginfo = NULL;
	}

	return res;

}

gboolean
gmyth_livetv_next_program_chain ( GMythLiveTV *livetv )
{
	gboolean res = TRUE;
	
	if ( !livetv->setup_done )
	{
		gmyth_debug ( "Call the setup function first!" );
		res= FALSE;
		goto error;		
	}
	
	/* Reload all TV chain from Mysql database. */
	gmyth_tvchain_reload_all (livetv->tvchain);

	if ( livetv->tvchain == NULL ) {
		res = FALSE;
		goto error;
	}

	// Get program info from database using chanid and starttime
	livetv->proginfo = gmyth_tvchain_get_program_at (livetv->tvchain, -1 );
	if ( livetv->proginfo == NULL ) {
		g_warning ("[%s] LiveTV not successfully started on the next program chain.\n", __FUNCTION__ );
		res = FALSE;
		goto error;
	} else {
		gmyth_debug ("[%s] GMythLiveTV: All requests to backend to start TV were OK, TV chain changed.\n", __FUNCTION__ );
	}

	return res;

error:
	g_print( "[%s] ERROR running LiveTV setup.\n", __FUNCTION__ );

	if ( livetv->local_hostname != NULL ) {
		g_string_free( livetv->local_hostname, TRUE );
		res = FALSE;
	}

	if ( livetv->recorder != NULL ) {
		g_object_unref (livetv->recorder);
		livetv->recorder = NULL;
	}

	if ( livetv->tvchain != NULL ) {
		g_object_unref (livetv->tvchain);
		livetv->tvchain = NULL;
	}

	if ( livetv->proginfo != NULL ) {
		g_object_unref (livetv->proginfo);
		livetv->proginfo = NULL;
	}

	return res;

}

GMythFileTransfer *
gmyth_livetv_create_file_transfer( GMythLiveTV *livetv )
{
	GMythURI* uri = NULL;
	
	if ( NULL == livetv || NULL == livetv->proginfo )
		goto done;
	
	if ( !livetv->setup_done )
	{
		gmyth_debug( "Error: You must do the LiveTV setup, just before generating the FileTransfer from LiveTV source!" );
		goto done;
	}	

  gmyth_debug( "URI path = %s.\n", livetv->proginfo->pathname->str ); 

	livetv->file_transfer = gmyth_file_transfer_new( livetv->backend_info );

  if ( NULL == livetv->file_transfer ) 
  {
  	gmyth_debug( "Error: couldn't create the FileTransfer from LiveTV source!" );
    goto done;
  }
  
  uri = gmyth_uri_new_with_value( livetv->proginfo->pathname->str );
  if ( NULL == uri )
  {
  	gmyth_debug( "Couldn't parse the URI to start LiveTV! [uri = %s]", livetv->proginfo->pathname->str );  	
  }

	if ( !gmyth_file_transfer_open( livetv->file_transfer, uri != NULL ? gmyth_uri_get_path(uri) : 
				livetv->proginfo->pathname->str ) )
	{
		gmyth_debug( "Error: couldn't open the FileTransfer from LiveTV source!" );
		g_object_unref( livetv->file_transfer );
		livetv->file_transfer = NULL;
		goto done;
	}

done:
	if ( uri != NULL )
	{
		g_object_unref( uri );
		uri = NULL;
	}	
	
	return livetv->file_transfer;
	
}

/* FIXME: How to proceed differently between livetv and recorded content */
void
gmyth_livetv_stop_playing (GMythLiveTV *livetv) 
{
	gmyth_debug ("[%s] Stopping the LiveTV...\n", __FUNCTION__);

	if (livetv->is_livetv) {
		if (!gmyth_recorder_stop_livetv (livetv->recorder)) {
			g_warning ("[%s] Error while stoping remote encoder", __FUNCTION__);	
		}
	}
}

// FIXME: put here just a wrapper function to call gmyth_recorder_is_recording()...
gboolean
gmyth_livetv_is_recording ( GMythLiveTV *livetv )
{
  gboolean ret = TRUE;

  GMythStringList *str_list = gmyth_string_list_new ();
  GString *message = g_string_new ("");

  gmyth_debug ( "[%s]\n", __FUNCTION__ );
  //g_static_mutex_lock (&mutex);
  //
  g_string_printf( message, "%s %d", GMYTHTV_RECORDER_HEADER,
      /* FIXME file_transfer->rec_id >= 0 ? file_transfer->rec_id : file_transfer->card_id );*/
        livetv->recorder->recorder_num);
  gmyth_string_list_append_string (str_list, message);
  gmyth_string_list_append_string (str_list, g_string_new ("IS_RECORDING"));

  gmyth_socket_sendreceive_stringlist ( livetv->recorder->myth_socket, str_list );

  if ( str_list != NULL && gmyth_string_list_length(str_list) > 0 )
  {
    GString *str = NULL;
    if ( ( str = gmyth_string_list_get_string( str_list, 0 ) ) != NULL && strcmp( str->str, "bad" )!= 0 ) {
      gint is_rec = gmyth_string_list_get_int( str_list, 0 );
      if ( is_rec != 0 )
        ret = TRUE;
      else
        ret = FALSE;
    }
  }
  g_print( "[%s] %s, stream is %s being recorded!\n", __FUNCTION__, ret ? "YES" : "NO", ret ? "" : "NOT" );
  //g_static_mutex_unlock (&mutex);

  if ( str_list != NULL )
    g_object_unref (str_list);

  return ret;

}


gboolean
gmyth_livetv_is_playing (GMythLiveTV *livetv)
{
	return TRUE;
}

void
gmyth_livetv_start_playing (GMythLiveTV *livetv)
{

	// TODO

}

